@@ -13,8 +13,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+along with this library; see the file COPYING.  If not, write to the Free
+Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 As a special exception, if you link this library with files
 compiled with a GNU compiler to produce an executable, this does not cause
@@ -40,7 +40,7 @@ the executable file might be covered by the GNU General Public License. */
  */
 
 /* Extensively hacked for GNU iostream by Per Bothner 1991, 1992, 1993.
-/* Changes copyright Per Bothner 1992, 1993. */
+   Changes copyright Free Software Foundation 1992, 1993. */
 
 #if defined(LIBC_SCCS) && !defined(lint)
 static char sccsid[] = "%W% (Berkeley) %G%";
@@ -74,6 +74,7 @@ static char sccsid[] = "%W% (Berkeley) %G%";
 #define	SUPPRESS	0x08	/* suppress assignment */
 #define	POINTER		0x10	/* weird %p pointer (`fake hex') */
 #define	NOSKIP		0x20	/* do not skip blanks */
+#define	WIDTH		0x40	/* width */
 
 /*
  * The following are used in numeric conversions only:
@@ -101,37 +102,39 @@ static char sccsid[] = "%W% (Berkeley) %G%";
 #define u_char unsigned char
 #define u_long unsigned long
 
-extern u_long strtoul _PARAMS((const char*, char**, int));
-extern u_long strtol _PARAMS((const char*, char**, int));
-static const u_char *__sccl _PARAMS((char *tab, const u_char *fmt));
-#ifndef USE_DTOA
+#ifdef __cplusplus
+extern "C" {
+#endif
+extern u_long strtoul __P((const char*, char**, int));
+extern long strtol __P((const char*, char**, int));
+static const u_char *__sccl __P((char *tab, const u_char *fmt));
+#ifndef _IO_USE_DTOA
 extern double atof();
 #endif
-
-#ifdef TODO
-/* If state is non-NULL, set failbit and/or eofbit as appropriate. */
+#ifdef __cplusplus
+}
 #endif
 
-int _IO_vfscanf(fp, fmt0, ap)
-     register _IO_FILE *fp;
-     char const *fmt0;
-     _IO_va_list ap;
-#ifdef TODO
-     ios *stream /* = NULL */;
-#endif
+/* If errp != NULL, *errp|=1 if we see a premature EOF;
+   *errp|=2 if we an invalid character. */
+
+int
+DEFUN(_IO_vfscanf, (fp, fmt0, ap, errp),
+      register _IO_FILE *fp AND char const *fmt0
+      AND _IO_va_list ap AND int *errp)
 {
 	register const u_char *fmt = (const u_char *)fmt0;
 	register int c;		/* character from format, or conversion */
 	register _IO_ssize_t width;	/* field width, or 0 */
 	register char *p;	/* points into all kinds of strings */
 	register int n;		/* handy integer */
-	register int flags;	/* flags as defined above */
+	register int flags = 0;	/* flags as defined above */
 	register char *p0;	/* saves original value of p when necessary */
 	int nassigned;		/* number of fields assigned */
 	int nread;		/* number of characters consumed from fp */
 	/* Assignments to base and ccfn are just to suppress warnings from gcc.*/
 	int base = 0;		/* base argument to strtol/strtoul */
-	typedef u_long (*strtoulfn) _PARAMS((const char*, char**, int));
+	typedef u_long (*strtoulfn) __P((const char*, char**, int));
 	strtoulfn ccfn = 0;
 	/* conversion function (strtol/strtoul) */
 	char ccltab[256];	/* character class table for %[...] */
@@ -151,8 +154,10 @@ int _IO_vfscanf(fp, fmt0, ap)
 		if (isspace(c)) {
 			for (;;) {
 			        c = _IO_getc(fp);
-				if (c == EOF)
-				    goto eof_failure;
+				if (c == EOF) {
+				  seen_eof++;
+				  break;
+				}
 				if (!isspace(c)) {
 				    _IO_ungetc (c, fp);
 				    break;
@@ -184,20 +189,26 @@ literal:
 			continue;
 
 		case '*':
-			flags |= SUPPRESS;
+			if (flags) goto control_failure;
+			flags = SUPPRESS;
 			goto again;
 		case 'l':
+			if (flags & ~(SUPPRESS | WIDTH)) goto control_failure;
 			flags |= LONG;
 			goto again;
 		case 'L':
+			if (flags & ~(SUPPRESS | WIDTH)) goto control_failure;
 			flags |= LONGDBL;
 			goto again;
 		case 'h':
+			if (flags & ~(SUPPRESS | WIDTH)) goto control_failure;
 			flags |= SHORT;
 			goto again;
 
 		case '0': case '1': case '2': case '3': case '4':
 		case '5': case '6': case '7': case '8': case '9':
+			if (flags & ~(SUPPRESS | WIDTH)) goto control_failure;
+			flags |= WIDTH;
 			width = width * 10 + c - '0';
 			goto again;
 
@@ -313,9 +324,9 @@ literal:
 		 * that suppress this.
 		 */
 		if ((flags & NOSKIP) == 0) {
-		    n = (unsigned char)*fp->_gptr;
+		    n = (unsigned char)*fp->_IO_read_ptr;
 		    while (isspace(n)) {
-			fp->_gptr++;
+			fp->_IO_read_ptr++;
 			nread++;
 			n = _IO_peekc(fp);
 			if (n == EOF)
@@ -338,11 +349,11 @@ literal:
 			if (flags & SUPPRESS) {
 			    _IO_size_t sum = 0;
 			    for (;;) {
-			        n = fp->_egptr - fp->_gptr;
+			        n = fp->_IO_read_end - fp->_IO_read_ptr;
 				if (n < (int)width) {
 				    sum += n;
 				    width -= n;
-				    fp->_gptr += n;
+				    fp->_IO_read_ptr += n;
 				    if (__underflow(fp) == EOF)
 					if (sum == 0)
 					    goto eof_failure;
@@ -352,16 +363,15 @@ literal:
 					}
 				} else {
 				    sum += width;
-				    fp->_gptr += width;
+				    fp->_IO_read_ptr += width;
 				    break;
 				}
 			    }
 			    nread += sum;
 			} else {
 			    _IO_size_t r =
-			      (*fp->_jumps->__xsgetn)(fp,
-						      (char*)va_arg(ap, char*),
-						      width);
+
+			      _IO_XSGETN (fp, (char*)va_arg(ap, char*), width);
 			    if (r != width)
 				goto eof_failure;
 			    nread += r;
@@ -376,8 +386,8 @@ literal:
 			/* take only those things in the class */
 			if (flags & SUPPRESS) {
 				n = 0;
-				while (ccltab[(unsigned char)*fp->_gptr]) {
-				    n++, fp->_gptr++;
+				while (ccltab[(unsigned char)*fp->_IO_read_ptr]) {
+				    n++, fp->_IO_read_ptr++;
 				    if (--width == 0)
 					break;
 				    if (_IO_peekc(fp) == EOF) {
@@ -391,8 +401,8 @@ literal:
 					goto match_failure;
 			} else {
 			    p0 = p = va_arg(ap, char *);
-			    while (ccltab[(unsigned char)*fp->_gptr]) {
-				*p++ = *fp->_gptr++;
+			    while (ccltab[(unsigned char)*fp->_IO_read_ptr]) {
+				*p++ = *fp->_IO_read_ptr++;
 				if (--width == 0)
 				    break;
 				if (_IO_peekc(fp) == EOF) {
@@ -417,8 +427,8 @@ literal:
 				width = ~0;
 			if (flags & SUPPRESS) {
 				n = 0;
-				while (!isspace((unsigned char)*fp->_gptr)) {
-					n++, fp->_gptr++;
+				while (!isspace((unsigned char)*fp->_IO_read_ptr)) {
+					n++, fp->_IO_read_ptr++;
 					if (--width == 0)
 						break;
 					if (_IO_peekc(fp) == EOF) {
@@ -429,8 +439,8 @@ literal:
 				nread += n;
 			} else {
 				p0 = p = va_arg(ap, char *);
-				while (!isspace((unsigned char)*fp->_gptr)) {
-					*p++ = *fp->_gptr++;
+				while (!isspace((unsigned char)*fp->_IO_read_ptr)) {
+					*p++ = *fp->_IO_read_ptr++;
 					if (--width == 0)
 						break;
 					if (_IO_peekc(fp) == EOF) {
@@ -450,7 +460,7 @@ literal:
 				width = sizeof(buf) - 1;
 			flags |= SIGNOK | NDIGITS | NZDIGITS;
 			for (p = buf; width; width--) {
-				c = (unsigned char)*fp->_gptr;
+				c = (unsigned char)*fp->_IO_read_ptr;
 				/*
 				 * Switch on the character; `goto ok'
 				 * if we accept it as a part of number.
@@ -534,7 +544,7 @@ literal:
 				 * c is legal: store it and look at the next.
 				 */
 				*p++ = c;
-				fp->_gptr++;
+				fp->_IO_read_ptr++;
 				if (_IO_peekc(fp) == EOF) {
 				    seen_eof++;
 				    break;		/* EOF */
@@ -581,7 +591,7 @@ literal:
 				width = sizeof(buf) - 1;
 			flags |= SIGNOK | NDIGITS | DPTOK | EXPOK;
 			for (p = buf; width; width--) {
-				c = (unsigned char)*fp->_gptr;
+				c = (unsigned char)*fp->_IO_read_ptr;
 				/*
 				 * This code mimicks the integer conversion
 				 * code, but is much simpler.
@@ -619,7 +629,7 @@ literal:
 				break;
 		fok:
 				*p++ = c;
-				fp->_gptr++;
+				fp->_IO_read_ptr++;
 				if (_IO_peekc(fp) == EOF) {
 				    seen_eof++;
 				    break;	/* EOF */
@@ -648,7 +658,7 @@ literal:
 			if ((flags & SUPPRESS) == 0) {
 				double res;
 				*p = 0;
-#ifdef USE_DTOA
+#ifdef _IO_USE_DTOA
 				res = _IO_strtod(buf, NULL);
 #else
 				res = atof(buf);
@@ -669,16 +679,13 @@ eof_failure:
 input_failure:
 	if (nassigned == 0)
 	    nassigned = -1;
+control_failure:
 match_failure:
-#ifdef TODO
-	if (stream)
-	    stream->set(ios::failbit);
-#endif
+	if (errp)
+	    *errp |= 2;
 done:
-#ifdef TODO
-	if (stream && seen_eof)
-		stream->set(ios::eofbit);
-#endif
+	if (errp && seen_eof)
+		*errp |= 1;
 	return (nassigned);
 }
 
@@ -688,9 +695,9 @@ done:
  * closing `]'.  The table has a 1 wherever characters should be
  * considered part of the scanset.
  */
-static const u_char *__sccl(tab, fmt)
-     register char *tab;
-     register const u_char *fmt;
+static const u_char *
+DEFUN(__sccl, (tab, fmt),
+      register char *tab AND register const u_char *fmt)
 {
 	register int c, n, v;
 
@@ -778,14 +785,3 @@ doswitch:
 	}
 	/* NOTREACHED */
 }
-
-#ifdef TODO
-int streambuf::scan(char const *format ...)
-{
-    va_list ap;
-    va_start(ap, format);
-    int count = vscan(format, ap);
-    va_end(ap);
-    return count;
-}
-#endif
